<?php

namespace Service;

use Util\Period;
use Util\Utilities;
use Util\Environment;
use Util\MathOperations;
use Util\ConditionsQuery;
use Zeedhi\Framework\DataSource\FilterCriteria;
use \Service\Manager\Mongo\Manager;
use Zeedhi\Zmart\DataSource\AggregateFunction;
use Zeedhi\Zmart\DataSource\ZmartFilterCriteria;
use Zeedhi\Zmart\QueryBuilders\Mongo\QueryBuilder;
use Service\Manager\Mongo\Types\Type;
use Zeedhi\Zmart\Util\ComparisonExpression;
use Zeedhi\Zmart\Util\LogicalExpression;
use Zeedhi\Zmart\Util\MathExpression;
use Zeedhi\Framework\DependencyInjection\InstanceManager;
use Util\PostGroupConditions;

class WidgetData {

    const TOTALIZED = "zmartTotalized";
    const GRID_CHART = 'grid';

    protected $dataSourceName = 'widget';
    /** @var \Service\ConnectionService $connectionService */
    protected $connectionService;
    /** @var \Service\Manager\Mongo\Manager $organizationManager */
    protected $organizationManager;
    /** @var DataSource $dataSourceService */
    protected $dataSourceService;
    /** @var \Service\Structure $structureService */
    protected $structureService;
    /** @var \Service\Indicator $indicatorService */
    protected $indicatorService;
    /** @var \Service\ViewType $viewTypeService */
    protected $viewTypeService;
    /** @var \Service\Hierarchy $hierarchyService */
    protected $hierarchyService;
    /** @var \Service\Hierarchy $hierarchyTypeService */
    protected $hierarchyTypeService;
    /** @var \Util\Environment $environment */
    protected $environment;
    /** @var \Util\Utilities $utilities */
    protected $utilities;

    protected $isViewPreview = false;
    protected $fullAccessProfiles = [];
    protected $isStructureFromFilter = false;

    public function __construct(ConnectionService $connectionService, DataSource $dataSourceService,
                                Structure $structure, Indicator $indicatorService,
                                Manager $organizationManager, ViewType $viewTypeService,
                                Hierarchy $hierarchyService, HierarchyType $hierarchyTypeService,
                                Environment $environment, Utilities $utilities) {
        $this->connectionService = $connectionService;
        $this->organizationManager = $organizationManager;
        $this->dataSourceService = $dataSourceService;
        $this->structureService = $structure;
        $this->indicatorService = $indicatorService;
        $this->viewTypeService = $viewTypeService;
        $this->hierarchyService = $hierarchyService;
        $this->hierarchyTypeService = $hierarchyTypeService;
        $this->environment = $environment;
        $this->utilities = $utilities;
    }

    /**
     * @param $id
     * @return array
     * @throws \Exception
     */
    public function findWidgetById($id) {
        $filterCriteria = new FilterCriteria($this->dataSourceName, array(
            array(
                "columnName" => "_id",
                "value" => $id,
                "operator" => "=")
            )
        );

        $widgets = $this->organizationManager->findBy($filterCriteria)->getRows();
        if(count($widgets) > 0) {
            return $widgets[0];
        }
        throw new \Exception(Exception::buildErrorJSON(
            "INDICATOR_NOT_FOUND",
            array("id" => $id)
        ), 1);
    }

    protected function getFactNames($facts) {
        $factNames = array();
        foreach ($facts as $fact) {
            $factNames[] = $fact['name'];
        }
        return $factNames;
    }

    public function getDataSourceColumns($dataSource, $columnName, $view = null) {
        $sourceColumns = $this->identifyColumnsToUse($dataSource, $view);
        $columns = array();

        if(!is_array($columnName)) {
            $columnName = $columnName ? array($columnName) : array();
        }
        if(!empty($columnName[0]) && is_string($columnName[0]) && isset($columnName[0]) && gettype(strpos($columnName[0], ',')) === "integer") {
            $columnName = explode(',', $columnName[0]);
        }

        if($sourceColumns) {
            foreach ($sourceColumns as $column) {
                if(in_array($column['name'], $columnName)) {
                    $columns[] = $column;
                }
            }
        }
        return $columns;
    }

    protected function identifyColumnsToUse($dataSource =  null, $view = null){
        $sourceColumns = array();

        $viewColumns = $view && isset($view['columns']) ? $view['columns'] : array();
        $dataSourceColumns = $dataSource && isset($dataSource['columns']) ? $dataSource['columns'] : array();

        foreach ($dataSourceColumns as $column) {
            $viewColumn = array_filter($viewColumns, function($vcolumn) use($column) {
                return $vcolumn['name'] === $column['name'];
            });

            if($viewColumn) {
                $viewColumn = reset($viewColumn);

                $column['class'] = isset($viewColumn['class']) ? $viewColumn['class'] : null;

                $properties = array('label', 'mask', 'direction', 'function');

                foreach ($properties as $property) {
                    if(isset($viewColumn[$property])){
                        $column[$property] = $viewColumn[$property];
                    }
                }
            }

            array_push($sourceColumns, $column);
        }

        return $sourceColumns;
    }

    protected function hasMultiplePeriod($period) {
        return is_array($period) && Utilities::isNumericArray($period);
    }

    public function getOrderBy(array $view, array $factColumns, array $dimensionColumns, $topFact = null) {
        $returnArr = array();

        if(isset($view['orderBy']) && !empty($view['orderBy'])) {
            if(is_array($view['orderBy'])){
                foreach ($view['orderBy'] as $viewOrderBy) {
                    $returnArr[$viewOrderBy['name']] = $viewOrderBy['direction'];
                }
            } else {
                $returnArr = array($view['orderBy'] => (isset($view['orderByDirection']) && !empty($view['orderByDirection'])) ? $view['orderByDirection'] : "ASC");
            }
        }

        foreach ($factColumns as $fact) {
            if(isset($fact['isSortable']) && ($fact['isSortable'] == true) && (empty($returnArr) || (!empty($topFact) && $topFact == $fact['name']))) {
                $returnArr = array($fact['name'] => (isset($fact['direction']) ? $fact['direction'] : "ASC"));
            }
        }

        foreach ($dimensionColumns as $dimension) {
            if(isset($dimension['isSortable']) && ($dimension['isSortable']) && empty($returnArr)) {
                $returnArr = array($dimension['name'] => (isset($dimension['direction']) ? $dimension['direction'] : "ASC"));
            }
        }

        return $returnArr;
    }

    public function buildGroupBy($dimension) {
        $groupBy = $dimension;
        $groupBy = ' GROUP BY '.implode(",",$groupBy);
        return $groupBy;
    }

    protected function getOrderByKeys($orderBy) {
        return implode(',', array_keys($orderBy));
    }

    protected function getDimensionsFromColumns ($columns) {
        $dimensions = [];
        foreach ($columns as $column) {
            if(!empty($column['isDimension']) && $column['isDimension'])
                $dimensions [] = $column['name'];
        }
        return $dimensions;
    }

    protected function getFactsFromColumns ($columns) {
        $facts = [];
        foreach ($columns as $column) {
            if(!empty($column['isFact']) && $column['isFact'])
                $facts [] = $column;
        }
        return $facts;
    }

    protected function buildSecondInlineView ($facts, $orderByArray, $columnFacts) {
        $returnArr = [];
        $factNames = $this->getFactNames($columnFacts);
        foreach ($facts as $fact) {
            $expression = (!Utilities::isColumnGenerated($fact)) ? $fact['expression'] : $fact['name'];
            $factFunction = isset($fact['function']) ? $fact['function'] : "";
            $returnArr [] = "{$factFunction}({$expression}) AS {$fact['name']}";

        }
        foreach ($orderByArray as $field) {
            foreach ($columnFacts as $fact) {
                $expression = (!Utilities::isColumnGenerated($fact)) ? $fact['expression'] : $fact['name'];
                $factFunction = isset($fact['function']) ? $fact['function'] : "";
                if($field == $fact['name']){
                    if(in_array($field, $factNames) && !in_array("{$factFunction}({$expression}) AS {$fact['name']}",$returnArr)){
                        $returnArr [] = "{$factFunction}({$expression}) AS {$fact['name']}";
                    }
                }
            }
        }
        return $returnArr;
    }

    protected function buildFirstInlineView ($facts, $orderByArray, $columnFacts) {
        $returnArr = [];
        foreach ($facts as $fact) {
            $returnArr [] = $fact['name'];
        }
        foreach ($columnFacts as $column) {
            if(in_array($column['name'], $orderByArray) &&
                !in_array($column['name'], $returnArr)){
                    $returnArr [] = $column['name'];
            }
        }
        return $returnArr;
    }

    protected function removeFactFromDimension ($columnFacts, $dimension) {
        $returnArr = [];
        $factNames = $this->getFactNames($columnFacts);
        foreach($dimension as $field){
            if(!in_array($field, $factNames)){
                $returnArr [] = $field;
            }
        }
        return $returnArr;
    }

    public function buildQuery($dataSource,$facts,&$dimension,$where,$orderby = array(),$groupProperties,$limitrows = null,
        $drillDownFilters = array(),$structure,$structureType,$groupByStructure,$columns,$isSpecialFunction) {

        $query = $dataSource['query'];
        $groupBy = '';
        $factNames = $this->getFactNames($facts);
        $firstInlineView = [];
        $secondInlineView = [];
        $orderBy = $this->getOrderByKeys($orderby);
        $orderByArray = explode(",", $orderBy);
        $columnDimensions = [];
        $columnFacts = [];
        $columnDimensions = $this->getDimensionsFromColumns($columns);
        $columnFacts = $this->getFactsFromColumns($columns);
        $secondInlineView = $this->buildSecondInlineView ($facts, $orderByArray, $columnFacts);
        $firstInlineView = $this->buildFirstInlineView ($facts, $orderByArray, $columnFacts);
        $dimension = $this->removeFactFromDimension($columnFacts,$dimension);

        $secondInlineView  = implode(',', $secondInlineView);
        $firstInlineView   = implode(',', $firstInlineView );
        $firstInlineView  .= ','.implode(',',$dimension);
        $secondInlineView .= ','.implode(',',$dimension);

        if (!empty($facts)) {
            if ($groupByStructure && ($structure || $structureType)) {
                $structure = is_array($structure) ? $structure[0] : $structure;
                $childStructureType = $structureType ? : $this->hierarchyTypeService->getChildStructureTypeByStructureType($structure['type']);
                $dimension = $this->getOracleGroupFromChildStructureType($dataSource, '', $childStructureType);
            }

            $dimension = !is_array($dimension) ? implode(",", $dimension) : $dimension;
            $groupBy = $this->buildGroupBy($dimension);
        } else {
            $selectQueryFields = ' * ';
        }

        $drillDownWhere = $this->buildQueryDrillDownWhere($drillDownFilters, $columns);
        $this->buildQueryWhere($where, $drillDownWhere);
        $query = <<<SQL
SELECT $firstInlineView FROM  ( SELECT $secondInlineView FROM ($query) A $where $groupBy)
SQL;

        if ($limitrows) {
            $query = 'SELECT * FROM (' . $query . ') A WHERE ROWNUM <= ' . $limitrows;
        }
        return $query;
    }


    protected function buildQueryOrderBy($facts, $orderBy, $dimension, $groupBy){
        $orderByString = "";
        $factNames = $this->getFactNames($facts);

        if(!empty($orderBy)) {
            foreach ($orderBy as $key => $direction) {
                if(!in_array($key, $factNames) && $key != $dimension) {
                    $groupBy .= empty($groupBy) ? " GROUP BY ". $key :  " ,".$key;
                }
                if(!empty($orderByString)){
                    $orderByString .= ", ";
                }
                $orderByString .= $key . " ". $direction;
            }
        }

        if(empty($orderByString)){
            return $groupBy;
        }

        return $groupBy." ORDER BY ".$orderByString;
    }

    protected function buildQueryDrillDownWhere($drillDownFilters, $columns){
        $drillDownWhere = "";
        if(!empty($drillDownFilters)) {
            foreach ($drillDownFilters as $key => $value) {
                if(is_string($value)){
                    $isDate = $this->findKeyColumn($columns, $key)['isDate'];
                    $this->buildDrillDrownWhere($drillDownWhere, $key, $value, $isDate);
                } else if(is_array($value)){
                    $newKeys = explode(',', $key);
                    for ($i = 0; $i < count($value); $i++) {
                        $newKey = $newKeys[$i];
                        $newValue = $value[$i];
                        $isDate = $this->findKeyColumn($columns, $newKey)['isDate'];
                        $this->buildDrillDrownWhere($drillDownWhere, $newKey, $newValue, $isDate);
                    }
                }
            }
        }

        return $drillDownWhere;
    }

    protected function findKeyColumn($columns, $key){
        $keyColumn = null;

        foreach ($columns as $column) {
            if($column['name'] === $key){
                $keyColumn = $column;
            }
        }

        return $keyColumn;
    }

    protected function buildDrillDrownWhere(&$drillDownWhere, $key, $value, $isDate){
        $keyValueComparsion = " {$key} = '{$value}' ";
        if($isDate){
            $keyValueComparsion = " {$key} = TO_DATE('{$value}') ";
        }
        $drillDownWhere .= empty($drillDownWhere)? $keyValueComparsion : " AND {$keyValueComparsion} ";
    }

    protected function buildQueryWhere(&$where, $drillDownWhere){
        if(!empty($where) && !empty($drillDownWhere)) {
            $where = " WHERE {$where} AND {$drillDownWhere}";
        } else if(!empty($where)) {
            $where = " WHERE {$where}";
        } else if(!empty($drillDownWhere)) {
            $where = " WHERE {$drillDownWhere}";
        }
    }

    protected function getOracleGroupFromChildStructureType($dataSource, $previousGroupBy, $childStructureType) {
        $groupBy = $previousGroupBy;

        if (isset($dataSource['relatedStructuresType'])) {
            $descriptionField = null;
            $columns = $this->identifyColumnsToUse($dataSource);
            foreach ($columns as $column) {
                if ($column['name'] == $childStructureType['descriptionField']) {
                    $descriptionField = $column['name'];
                }
            }
            if ($descriptionField) {
                $groupBy .= empty($groupBy)? $descriptionField : ", {$descriptionField}";
            } else {
                foreach ($dataSource['relatedStructuresType'] as $relatedStructureType) {
                    if($relatedStructureType['structureType'] === $childStructureType[Manager::MONGO_KEY_COLUMN]) {
                        foreach ($relatedStructureType['relations'] as $relation) {
                            if (in_array($relation['structureColumn'], $childStructureType['fieldKey'])) {
                                $descriptionField = $relation['dataSourceColumn'];
                                $groupBy .= empty($groupBy)? "{$descriptionField} " : ", {$descriptionField} ";
                            }
                        }
                    }
                }
            }
        }

        return $groupBy;
    }

    public function fetchWidgetData($viewId, $structureId = null, $structureTypeId = null, $period = null,
                                    $facts = array(), $dimension = null, $orderBy = array(), $groupBy = null,
                                    $limitRows = null, $drillDownFilters = array(), $groupByStructure = false,
                                    $previousConditions = array(), $postConditions = array(), $postGroupConditions=[],
                                    $filterOrderByTotal = null,
                                    $options = array(), $dimensionX = null, $dimensionY = null, $isViewPreview = false,
                                    $fullAccessProfiles = [], $isStructureFromFilter = false)
    {
        $accessProfiles = Utilities::getUtilities()->getAccessProfiles();
        if(!empty($viewId) && (!empty($structureId) || !empty($structureTypeId) || Utilities::arrayIntersects($accessProfiles,$fullAccessProfiles)) || !empty($isViewPreview)){
            if($isViewPreview){
                $importDatabase = Utilities::getImportDefaultDatabase();
                $this->organizationManager->createDatabase($importDatabase);
                $this->isViewPreview = true;
            } else {
                $this->isViewPreview = false;
            }

            $this->fullAccessProfiles = $fullAccessProfiles;
            $this->isStructureFromFilter = $isStructureFromFilter;
            $view = $this->findWidgetById($viewId);
            $serieStacked = isset($view['serieStacked']) ? $view['serieStacked'] : null;
            if (isset($view['dateFilterColumn'])) {
                $options['dateFilterColumn'] = $view['dateFilterColumn'];
            }
            $viewTypeId = $view['viewTypeId'];
            $indicatorId = $view["indicator"];
            $structure = null;
            $structureType = null;
            $viewDimension = $dimension;

            if (!$limitRows && isset($view['limitRows'])) {
                $limitRows = $view['limitRows'];
            }

            if($view['template'] === 'zmart-components#compare-periods.html') {
                $limitRows = null;
            }

            $viewType = $this->viewTypeService->findById($viewTypeId);

            if (!$viewDimension) {
                if (isset($viewType['dimensionProperty']) && isset($viewType['dimensionProperty']['name'])) {
                    $dimensionProperty = $viewType['dimensionProperty']['name'];
                    if (isset($view[$dimensionProperty])) {
                        $viewDimension = $view[$dimensionProperty];
                    }
                }
            }

            if ($groupByStructure && $structureTypeId) {
                $structureType = $this->structureService->findStructureTypeById($structureTypeId);
            }

            $structures = array();

            if ($structureId) {
                $structures = $this->structureService->findByIds($structureId);
            }
            $dataSource = $this->dataSourceService->findByIndicator($indicatorId);
            $connectionId = isset($dataSource['connection']) ? $dataSource['connection'] : null;
            if(empty($facts)) {
                $facts = $this->getDefaultFacts($view, $viewType);
            }else{
                $this->gaugeFacts($view,$facts);
            }
            $factsNames = $facts;
            $facts = $this->getDataSourceColumns($dataSource, $facts, $view);
            $haveRelatedFact = false;
            $hasSpecialFunction = false;
            foreach($facts as $fact){
                if(!empty($fact["relatedFactMultiple"])){
                    foreach($fact["relatedFactMultiple"] as $fact){
                        if(!array_search($fact,$factsNames)){
                            array_push($factsNames,$fact);
                            $haveRelatedFact = true;
                        }
                    }
                }
                if(isset($fact["relatedFact"]) && !array_search($fact["relatedFact"],$factsNames)){
                    array_push($factsNames,$fact["relatedFact"]);
                    $haveRelatedFact = true;
                }
                if(!empty($fact["specialFunction"])){
                    $hasSpecialFunction = true;
                }
            }
            if($haveRelatedFact){
                $facts = $factsNames;
                $facts = $this->getDataSourceColumns($dataSource, $facts, $view);
            }
            if(!$dimension) {
                $dimension = $this->getDefaultDimension($view, $viewType);
            }
            $dimensions = $this->getDataSourceColumns($dataSource, $dimension, $view);
            if(empty($orderBy) && !empty($facts) && !empty($dimensions)) {
                $topFact = null;
                if(!empty($viewType["name"]) && $viewType["name"] == "dashboardCard" && !empty($view["cardValue"])){
                    $topFact = $view["cardValue"];
                }
                $orderBy = $this->getOrderBy($view, $facts, $dimensions, $topFact);
            }

            if (is_null($orderBy)) {
                $orderBy = array();
            }

            $specialFunctionColumns = $this->getSpecialFunctionColumns($view['columns']);
            $isSpecialFunction = false;
            foreach ($orderBy as $columnName => $direction) {
                if(!is_array($dimension)){
                    $dimension = array($dimension);
                }
                if(!in_array($columnName, $specialFunctionColumns)){
                    if((is_array($factsNames) && !in_array($columnName, $factsNames) && !in_array($columnName, $dimension))
                    || (is_string($factsNames) && $columnName !== $factsNames && !in_array($columnName, $dimension))) {
                        if(is_array($dimension)) {
                            $dimension[] = $columnName;
                        }
                        else {
                            $dimension = array($dimension, $columnName);
                        }
                    }
                } else {
                    $isSpecialFunction = true;
                }
            }

            if($period && $this->hasMultiplePeriod($period)) {
                $results = array();
                foreach ($period as $singlePeriod) {
                    if($singlePeriod != null){
                        $singlePeriod = $this->decodePeriod($singlePeriod);
                        $suffix = implode("_", Period::formatInterval(array($singlePeriod['start'], $singlePeriod['end'])));
                        $data = $this->getDataFromParams($structures, $structureType, $dataSource, $facts, $dimension,
                            $viewDimension, $orderBy, $groupBy, $singlePeriod, $limitRows, $drillDownFilters, $connectionId,
                            $groupByStructure, $previousConditions, $postConditions, $options, $view);
                        $results[] = $this->modifyFactNameByPeriod($suffix, $factsNames, $data);
                    }
                    else{
                        $results[] = $this->getDataFromParams($structures, $structureType, $dataSource, $facts, $dimension,
                            $viewDimension, $orderBy, $groupBy, $singlePeriod, $limitRows, $drillDownFilters, $connectionId,
                            $groupByStructure, $previousConditions, $postConditions, $options, $view);
                    }
                }

                if(!isset($view['serieStacked']))
                {
                    $view['serieStacked'] = "";
                }
                $returnData = $this->groupByProperty($dimension, $results);
                $returnData = $this->sortData($returnData, $orderBy, $serieStacked, $filterOrderByTotal, $view, 'ID',$factsNames);
            } else {
                if ($period){
                    $period = $this->decodePeriod($period);
                }
                $returnData = $this->getDataFromParams($structures, $structureType, $dataSource, $facts,
                    $dimension, $viewDimension, $orderBy, $groupBy, $period, $limitRows, $drillDownFilters,
                    $connectionId, $groupByStructure, $previousConditions, $postConditions, $options, $view,
                    $isSpecialFunction);
                $returnData = $this->sortData($returnData, $orderBy, $serieStacked, $filterOrderByTotal, $view, 'ID',$factsNames);
            }

            if($view['serieType'] == 'crosstaba') {
                if(isset($facts[0]['name'])){
                    $factValue = $facts[0]['name'];
                }else{
                    $factValue = $view['yKeys'][0];
                }
                if($dimensionY){
                    $view['yAxisProperty'] = $dimensionY;
                }
                if(!is_array($view['yAxisProperty'])){
                    $view['yAxisProperty'] = array($view['yAxisProperty']);
                }
                $dimensions = $view['yAxisProperty'];
                if($dimensionX){
                    $view['xAxisProperty'] = $dimensionX;
                }
                $factLabel = $view['xAxisProperty'];
                $returnData = $this->setDataSetInCrossTab($factLabel,$factValue,$dimensions,$returnData);
            }
            if($hasSpecialFunction && !$dataSource["hasDataMining"]){
                foreach($facts as $fact){
                    if(!empty($fact['specialFunction'])){
                        if($fact["specialFunction"]=="totalPercent"){
                            $returnData = $this->totalPercent($returnData, $facts, $fact, $structures, $structureType, $dataSource, $orderBy, $period, $limitRows, $drillDownFilters, $groupByStructure, $previousConditions, $postConditions, $options, $view);
                        }else{
                            foreach($returnData as &$data){
                                $relatedFact = !empty($data[$fact["relatedFact"]]) ? $data[$fact["relatedFact"]] : null;
                                $data[$fact["name"]] = $this->getValueFromSpecialFunction($fact['specialFunction'],$relatedFact, $fact, $data, false);
                            }
                        }
                    }
                }
            }
            if(!empty($postGroupConditions)){
                $returnData = PostGroupConditions::filter($returnData, $postGroupConditions);
            }
            return $returnData;
        }
        return [];
    }

    protected function factTotal($returnData, $factName){
        $total = 0;
        foreach($returnData as $data){
            $total += !empty($data[$factName]) ? $data[$factName] : 0;
        }
        return $total;
    }

    protected function sortData($returnData, $orderBy, $serieStacked, $filterOrderByTotal, $view, $key,$factsNames) {
        if (!empty($orderBy)) {
            $key = array_keys($orderBy);
            if(!(sizeof($key) > 1))
            {
                $key = implode("|",$key);
            }
            $orderByString = implode("|", $orderBy);
            $sumTotal = 0;


            if (is_null($filterOrderByTotal) || is_null($serieStacked)) {
                if (isset($view['serieStacked']) && $view['serieStacked']) {
                    if (isset($view['orderByTotal']) && $view['orderByTotal'] && $view['chartSeries'][0]['serieStacked']) {
                        foreach ($returnData as $index => $data) {
                            foreach ($factsNames as $name) {
                                $sumTotal += $data[$name];
                            }
                            $returnData[$index]['sumTotal'] = $sumTotal;
                            $sumTotal = 0;
                        }
                        $returnData = Utilities::getUtilities()->orderUsort($returnData, 'sumTotal', $orderByString, $view['columns']);
                    } else {
                        if (!is_string($key) && sizeof($key) > 1) {
                            $returnData = Utilities::getUtilities()->orderMultiSort($returnData, $key, explode('|', $orderByString), $view['columns']);
                        } else {
                            $returnData = Utilities::getUtilities()->orderUsort($returnData,$key,$orderByString,$view['columns']);
                        }
                    }
                } else {
                    if (!is_string($key) && sizeof($key) > 1) {
                        $returnData = Utilities::getUtilities()->orderMultiSort($returnData, $key, explode('|', $orderByString), $view['columns']);
                    } else {
                        $returnData = Utilities::getUtilities()->orderUsort($returnData, $key, $orderByString, $view['columns']);
                    }
                }
            } else {
                if (isset($serieStacked) && $serieStacked) {
                    if (isset($filterOrderByTotal) && $filterOrderByTotal) {
                        foreach ($returnData as $index => $data) {
                            foreach ($factsNames as $name) {
                                $sumTotal += $data[$name];
                            }
                            $returnData[$index]['sumTotal'] = $sumTotal;
                            $sumTotal = 0;
                        }
                        $returnData = Utilities::getUtilities()->orderUsort($returnData, 'sumTotal', $orderByString, $view['columns']);
                    } else {
                        if (!is_string($key) && sizeof($key) > 1) {
                            $returnData = Utilities::getUtilities()->orderMultiSort($returnData, $key, explode('|', $orderByString), $view['columns']);
                        } else {
                            $returnData = Utilities::getUtilities()->orderUsort($returnData, $key, $orderByString, $view['columns']);
                        }
                    }
                } else {
                    if (!is_string($key) && sizeof($key) > 1) {
                        $returnData = Utilities::getUtilities()->orderMultiSort($returnData, $key, explode('|', $orderByString), $view['columns']);
                    } else {
                        $returnData = Utilities::getUtilities()->orderUsort($returnData, $key, $orderByString, $view['columns']);
                    }
                }
            }
        }
        if($view["serieType"]=="pie"){
            $others=0;
            $x=$view["chartXAxisField"];
            if(is_array($returnData) && count($returnData) >= 20){
                foreach($returnData as $i=>$r){
                    if($i>=9){
                        $others+=$r[$key];
                        unset($returnData[$i]);
                    }
                }
                $returnData[9]=array($x => "Others", $key=>$others);
            }
        }
        return $returnData;
    }

    protected function getSpecialFunctionColumns ($columns) {
        foreach($columns as $column){
            if(isset($column['specialFunction']) && $column['specialFunction'] == "case"){
                $specialFunctionColumns[] = $column['name'];
            } else {
                $specialFunctionColumns[] = "";
            }
        }
        return $specialFunctionColumns;
    }

    protected function decodePeriod($period) {
        $periodToReturn = array();
        if(is_string($period)) {
            $periodName = $period;
            list($periodToReturn['start'], $periodToReturn['end']) = Period::decode($periodName);
        } else if($period && isset($period['start'])) {
            list($periodToReturn['start'], $periodToReturn['end']) = Period::convertStringToMoment($period['start'], $period['end']);
        }
        return $periodToReturn;
    }

    public function modifyFactNameByPeriod($suffix, $facts, $data){
        return array_map(function($item) use ($suffix, $facts) {
            foreach ($facts as $fact) {
                if (isset($item[$fact])) {
                    $item["{$fact}_{$suffix}"] = $item[$fact];
                    $item["suffix"] = $suffix;
                    unset($item[$fact]);
                }
            }
            return $item;
        }, $data);
    }

    protected function groupByProperty($dimension, $results) {
        $groupedData = array();
        if(is_array($dimension)) {
            foreach ($results as &$result) {
                foreach ($result as &$item) {
                    $customDimension = array();
                    foreach($dimension as $singleDimension) {
                        if(isset($item[$singleDimension])) {
                            $customDimension[] = $item[$singleDimension];
                        }
                    }
                    $customDimension = implode('_', $customDimension);
                    $item['zmartCustomDimension'] = $customDimension;
                }
            }
            $dimension = "zmartCustomDimension";
        }

        if(is_array($dimension)) {
            $dimension = implode(',', $dimension);
        }

        foreach ($results as $resultData) {
            foreach ($resultData as $data) {
                if (!isset($groupedData[$data[$dimension]])) {
                    $groupedData[$data[$dimension]] = $data;
                } else {
                    $groupedData[$data[$dimension]] = array_merge($groupedData[$data[$dimension]], $data);
                }
            }
        }

        return array_values($groupedData);
    }

    protected function getDataFromParams($structures, $structureType, $dataSource, $facts,
        $dimension, $viewDimension, $orderBy, $groupBy, $period, $limitRows, $drillDownFilters,
        $connectionId = null, $groupByStructure, $previousViewConditions, $postViewConditions,
        $options = array(), $view, $isSpecialFunction = null) {

        $columns = $this->identifyColumnsToUse($dataSource, $view);

        if(isset($dataSource['hasDataMining']) && $dataSource['hasDataMining']) {
            $returnData = $this->getDataMiningData($structures, $structureType, $dataSource,
                $facts, $dimension, $orderBy, $period, $limitRows, $drillDownFilters,
                $groupByStructure, $previousViewConditions, $postViewConditions, $options, $view);
        } else {
            $returnData = $this->getOracleData($structures, $structureType, $dataSource, $facts,
                $dimension, $viewDimension, $orderBy, $groupBy, $period, $limitRows,
                $drillDownFilters, $connectionId, $groupByStructure,$previousViewConditions,
                $postViewConditions, $columns, $options, $isSpecialFunction, $view);
        }
        return $returnData;
    }

    protected function getDataMiningData($structures, $structureType, $dataSource, $facts,
        $dimension, $orderBy, $period, $limitRows, $drillDownFilters, $groupByStructure,
        $previousViewConditions, $postViewConditions, $options = array(), $view)
    {
        $factsWithSpecialFunctions = $this->getFactsWithSpecialFunction($facts);
        if(!empty($factsWithSpecialFunctions)) {
            $facts = $this->addSpecialFunctionRelatedFacts($factsWithSpecialFunctions, $facts, $dataSource['columns']);
        }

        $filterCriteria = $this->mapFilterCriteria($structures, $structureType, $dataSource, $facts, $dimension, $orderBy,
            $period, $limitRows, $drillDownFilters, $groupByStructure, $previousViewConditions, $postViewConditions,
            $options, $view);

        $allowDivisionByZero = isset($view['allowDivisionByZero']) ? $view['allowDivisionByZero'] : false;
        $returnData = $this->organizationManager->aggregate($filterCriteria, $allowDivisionByZero)->getRows();

        if($options && isset($options[self::TOTALIZED]) && $options[self::TOTALIZED]){
            $returnData = $this->getTotalized($view, $returnData, $structures);
        }

        if(!empty($factsWithSpecialFunctions)) {

            $returnData = $this->calcSpecialFunctions($factsWithSpecialFunctions, $returnData, $facts, $structures,
                $structureType, $dataSource, $orderBy, $period, $limitRows, $drillDownFilters, $groupByStructure,
                $previousViewConditions, $postViewConditions, $options, $view);
        }

        return $this->prepareDataForResult($returnData, $dimension, $facts);
    }

    protected function addSpecialFunctionRelatedFacts($factsWithSpecialFunctions, $facts, $columns) {
        $specialFunctionsToConsider = array('abs', 'sign', 'ceil', 'floor', 'sqrt', 'mod', 'power', 'case');

        foreach ($factsWithSpecialFunctions as $factWithSpecialFunction) {
            $specialFunction = $factWithSpecialFunction['specialFunction'];
            if(in_array($specialFunction, $specialFunctionsToConsider)) {
                $facts = $this->addRelatedFactToArray($columns, $factWithSpecialFunction['relatedFact'] ?? "", $facts);
            }

            if($specialFunction === 'mod') {
                $facts = $this->addRelatedFactToArray($columns, $factWithSpecialFunction['secondRelatedFact'] ?? "", $facts);
            }
        }

        return $facts;
    }

    protected function addRelatedFactToArray($columns, $factName, $facts) {
        $relatedFact = $this->searchByProperty($columns, $factName, "name");
        $isFactAlreadyRelated = $this->searchByProperty($facts, $relatedFact['name'], 'name');
        if(!$isFactAlreadyRelated) {
            array_push($facts, $relatedFact);
        }

        return $facts;
    }
    protected static function getNameStructuresTotalized($structures){
        $structures = Utilities::getArrayFromAssociativeArray($structures,"name");
        $nameStructures = Utilities::getStringForArray($structures);
        return $nameStructures;
    }

    protected function getTotalized($view, $returnData, $structures) {
        $nameStructures = self::getNameStructuresTotalized($structures);
        if($view['viewType'] === self::GRID_CHART){
            foreach ($returnData as &$data) {
                $data["STRUCTURE"] = $nameStructures;
            }
        } else {
            $returnData = $this->calcTotal($returnData);
        }

        return $returnData;
    }

    protected function calcTotal($data) {
        $totalRow = array();
        foreach($data as $row) {
            foreach($row as $property => $propertyValue) {
                if($property !== '_id' && is_numeric($propertyValue)) {
                    if(!isset($totalRow[$property])) {
                        $totalRow[$property] = 0;
                    }
                    $totalRow[$property] += floatval($propertyValue);
                }
            }
        }
        $totalRow[QueryBuilder::MONGO_ID] = "Total";
        $data[] = $totalRow;
        return $data;
    }

    protected function calcSpecialFunctions($factsWithSpecialFunctions, $returnData, $facts, $structures, $structureType, $dataSource, $orderBy, $period, $limitRows, $drillDownFilters, $groupByStructure, $previousViewConditions, $postViewConditions, $options, $view) {
        foreach ($factsWithSpecialFunctions as $factWithSpecialFunction) {
            $function = $factWithSpecialFunction['specialFunction'];

            if($function === 'totalPercent') {
                $returnData = $this->totalPercent($returnData, $facts, $factWithSpecialFunction, $structures, $structureType, $dataSource, $orderBy, $period, $limitRows, $drillDownFilters, $groupByStructure, $previousViewConditions, $postViewConditions, $options, $view);
            } else {
                foreach ($returnData as &$row) {
                    $relatedFactName = isset($factWithSpecialFunction['relatedFact'])?$factWithSpecialFunction['relatedFact']:null;
                    $relatedFactValue = isset($row[$relatedFactName]) ? $row[$relatedFactName] : null;
                    $row[$factWithSpecialFunction['name']] = $this->getValueFromSpecialFunction($function, $relatedFactValue, $factWithSpecialFunction, $row);
                }
            }
        }

        return $returnData;
    }

    protected function totalPercent($returnData, $facts, $factWithSpecialFunction, $structures, $structureType, $dataSource, $orderBy, $period, $limitRows, $drillDownFilters, $groupByStructure, $previousViewConditions, $postViewConditions, $options, $view) {
        $customDimension = "zmartNullDimension";
        $relatedFact = $this->searchByProperty($facts, $factWithSpecialFunction['relatedFact'], "name");
        $customFilter = $this->mapFilterCriteria($structures, $structureType, $dataSource, array($relatedFact),  $customDimension, $orderBy, $period, $limitRows, $drillDownFilters, $groupByStructure, $previousViewConditions, $postViewConditions, $options, $view);
        $aggregationTotal = $this->organizationManager->aggregate($customFilter)->getRows();
        $totalValue = isset($aggregationTotal[0][$relatedFact['name']]) ? $aggregationTotal[0][$relatedFact['name']] : $this->factTotal($returnData, $relatedFact["name"]);

        foreach($returnData as &$row) {
            if(isset($row[$relatedFact['name']])) {
                $row[$factWithSpecialFunction['name']] = ($row[$relatedFact['name']] / $totalValue) * 100;
            }
        }

        return $returnData;
    }

    protected function getValueFromSpecialFunction($function, $relatedFactValue, $factWithSpecialFunction, $row, $isDataMining = true) {
        switch ($function) {
            case 'abs':
                return abs($relatedFactValue);
            case 'sign':
                return MathOperations::sign($relatedFactValue);
            case 'ceil':
                return ceil($relatedFactValue);
            case 'floor':
                return floor($relatedFactValue);
            case 'sqrt':
                return sqrt($relatedFactValue);
            case 'mod':
                return $this->calcModFunction($factWithSpecialFunction, $row,$relatedFactValue);
            case 'power':
                $expoent = isset($factWithSpecialFunction['potentialExpoent']) ? $factWithSpecialFunction['potentialExpoent'] : 1;
                return pow($relatedFactValue, $expoent);
            case 'case':
                return $this->getValueSpecialFunctionCase($isDataMining, $factWithSpecialFunction, $row);
            default:
                return null;
        }
    }

    protected function getValueSpecialFunctionCase($isDataMining, $factWithSpecialFunction, $row){
        return $isDataMining ? $this->calcCaseFunction($factWithSpecialFunction, $row) : $row[$factWithSpecialFunction["name"]];
    }

    protected function calcCaseFunction($factWithSpecialFunction, $row) {
        $expression = explode('DEFAULT ', !empty($factWithSpecialFunction['caseExp']) ? $factWithSpecialFunction['caseExp'] : $factWithSpecialFunction['expression']);
        $caseExpressions = explode('CASE ', $expression[0]);
        $defaultValueExpression = $this->utilities->replaceParamsToEvaluate($expression[1]);

        foreach ($caseExpressions as $caseExpression) {
            if($caseExpression) {
                $caseExpression = explode('THEN ', $caseExpression);
                $condition = $this->utilities->replaceParamsToEvaluate($caseExpression[0]);
                $value = $this->utilities->replaceParamsToEvaluate($caseExpression[1]);
                if(eval('return '.$condition.';')) {
                    return eval('return '.$value.';');
                }
            }
        }
        return eval('return '.$defaultValueExpression.';');
    }

    protected function calcModFunction($factWithSpecialFunction, $row, $relatedFactValue) {
        $secondRelatedFactName = !empty($factWithSpecialFunction['secondRelatedFact']) ? $factWithSpecialFunction['secondRelatedFact'] : "";
        $secondRelatedFactValue = isset($row[$secondRelatedFactName]) ? $row[$secondRelatedFactName] : null;

        $modSecondValue = isset($factWithSpecialFunction['modSecondValue']) ? $factWithSpecialFunction['modSecondValue'] : null;
        $modSecondValue = !empty($secondRelatedFactValue) ? $secondRelatedFactValue : $modSecondValue;
        if(!$modSecondValue){
            $modSecondValue = $relatedFactValue+1;
        }

        return $relatedFactValue % $modSecondValue;
    }

    protected function searchByProperty($hayStack, $needle, $property) {
        $elementToReturn = null;
        foreach($hayStack as $hay) {
            $elementToReturn = ($hay[$property] === $needle) ? $hay : $elementToReturn;
        }
        return $elementToReturn;
    }

    protected function getFactsWithSpecialFunction (&$facts) {
        $factsWithSpecialFunctions = array();
        foreach($facts as $index => $fact) {
            if($this->hasFactSpecialFunction($fact)) {
                $factsWithSpecialFunctions[] = $fact;
                unset($facts[$index]);
            }
        }
        return $factsWithSpecialFunctions;
    }

    protected function hasFactSpecialFunction($fact) {
        return isset($fact['specialFunction']) && !!$fact['specialFunction'];
    }

    protected function getOracleData($structures, $structureType, $dataSource, $facts,
        $dimension, $viewDimension, $orderBy, $groupBy, $period, $limitRows,
        $drillDownFilters, $connectionId = null, $groupByStructure,$previousViewConditions,
        $postViewConditions, $columns, $options, $isSpecialFunction, $view)
    {
        if(isset($period['start']) && isset($period['end'])) {
            $period = array($period['start'], $period['end']);
        }

        $previousConditionsExpression = !empty($previousViewConditions) ? ConditionsQuery::convertLogicalToSqlSentence($previousViewConditions) : "" ;
        $postConditionsExpression = !empty($postViewConditions) ? ConditionsQuery::convertLogicalToSqlSentence($postViewConditions) : "";
        $where = $this->dataSourceService->buildWhereClause($structures, $dataSource, $previousConditionsExpression);
        $query = $this->buildQuery($dataSource, $facts, $dimension, $where, $orderBy, $groupBy, $limitRows, $drillDownFilters,
            $structures, $structureType, $groupByStructure, $columns, $isSpecialFunction);
        if($period) {
            list($period['start'], $period['end']) = Period::formatInterval($period);
            $periodConditions = array(
                array("columnName" => "period_start", "operator" => FilterCriteria::EQ, "value" => $period['start']),
                array("columnName" => "period_end", "operator" => FilterCriteria::EQ, "value" => $period['end'])
            );
        } else {
            $periodConditions = array();
        }

        $filterCriteria = new FilterCriteria($this->dataSourceName, $periodConditions);
        $organizationId = Utilities::getOrganizationIdForMongo();
        $clientRoom = $this->connectionService->getClientRoom($organizationId, $dataSource);

        $dataSourceName = 'queryRunny';
        if(isset($dataSource['name'])){
            $dataSourceName = $dataSource['name'];
        }

        $data = $this->dataSourceService->execQueryFromWebSocket($query, $filterCriteria, $clientRoom, 60,
            $connectionId, $postConditionsExpression, $orderBy, true,$dataSourceName);
        if ($groupByStructure) {
            $data = $this->prepareOracleDataForResult($data, $dimension, $viewDimension);
        }

        if($options && isset($options[self::TOTALIZED]) && $options[self::TOTALIZED]){
            $data = $this->getTotalized($view, $data, $structures);
        }

        $data = $this->convertDateData($data, $columns);
        return $data;
    }

    protected function convertDateData($data, $columns){
        foreach ($data as &$dataSet) {
            foreach ($dataSet as $key => &$value) {
                $column = $this->findKeyColumn($columns, $key);
                if($column['isDate']){
                    if(is_string($value) &&
                        strlen($value) == 19 &&
                        substr_count($value, '/') == 2 ) {
                        $dateValue = \DateTime::createFromFormat('d/m/Y H:i:s',$value);
                    }else{
                        $dateValue = new \DateTime($value);

                    }
                    $value = $dateValue->format('d/m/Y');
                }
            }
        }

        return $data;
    }

    protected function prepareOracleDataForResult($data, $dimensions, $viewDimension) {
        $dimensions = explode(',', $dimensions);
        if (is_string($viewDimension)) {
            foreach ($data as &$dataItem) {
                foreach ($dimensions as $dimension) {
                    if (array_key_exists($dimension, $dataItem)) {
                        $value = $dataItem[$dimension];
                        unset($dataItem[$dimension]);
                        $dataItem[$viewDimension] = $value?: '';
                    }
                }
            }
        }

        return $data;
    }

    protected function prepareDataForResult($data, $dimension, $facts) {
        foreach ($facts as $fact) {
            if(isset($fact['isFunctionAccumulative']) && $fact['isFunctionAccumulative']) {
                if(!empty($fact["mathExpressions"])){
                    $data = Accumulate::accumulate($data, $fact['name']);
                }else{
                    $data = Accumulate::accumulate($data, $fact['relatedFact'], $fact['name']);
                }
            }else if(!empty($fact['relatedFact']) && empty($fact["mathExpressions"])){
                foreach($data as &$d){
                    $d[$fact['name']] = $d[$fact['relatedFact']];
                }
            }
        }
        if(is_array($dimension)) {
            $dimension = $dimension[0];
        }
        foreach ($data as &$dataItem) {
            $label = $dataItem[QueryBuilder::MONGO_ID];
            unset($dataItem[QueryBuilder::MONGO_ID]);

            if(is_string($label) || is_null($label)) {
                $dataItem[$dimension] = $label;
            } else if($label instanceof \DateTime) {
                $dataItem[$dimension] = $label->format('d/m/Y');
            } else if(is_array($label)) {
                $label = $this->processDimensionLabel($label);
                $dataItem = array_merge($dataItem, $label);
            }
        }
        return $data;
    }

    protected function processDimensionLabel(array $label) {
        foreach($label as &$labelProperty) {
            if($labelProperty instanceof \DateTime){
                $labelProperty = $labelProperty->format('d/m/Y');
            }
        }
        return $label;
    }

    protected function mapFilterCriteria($structures, $structureType, $dataSource, $facts,
        $dimension, $orderBy, $period, $limitRows, $drillDownFilters, $groupByStructure,
        $previousViewConditions, $postViewConditions, $options, $view)
    {
        $dataSourceId = $dataSource[QueryBuilder::MONGO_ID];
        $filterCriteria = new ZmartFilterCriteria($dataSourceId, array(), null, $limitRows > 0 ? $limitRows : 0);
        $this->setFilterCriteriaWhere($filterCriteria, $dataSource, $period, $structures, $drillDownFilters, $previousViewConditions, $postViewConditions, $options, $view);
        $this->setFilterCriteriaAggregation($filterCriteria, $facts, $dimension, $orderBy, $dataSource, $structures, $structureType, $groupByStructure, $view);

        return $filterCriteria;
    }

    protected function setFilterCriteriaWhere(ZmartFilterCriteria $filterCriteria, $dataSource,
        $period, $structures, $drillDownFilters, $previousViewConditions, $postViewConditions,
         $options, $view)
    {
        $dateFilterColumn = isset($options['dateFilterColumn'])? $options['dateFilterColumn'] : null;
        if($period){
            $this->addWhereDateRangeFromPeriod($filterCriteria, $dataSource, $period, $dateFilterColumn);
        }

        $this->addWhereFromRelations($filterCriteria, $dataSource, $structures);
        $this->addWhereFromDrillDownfilters($filterCriteria, $drillDownFilters, $dataSource, $view);
        $this->addWhereFromCustomViewConditions($filterCriteria, $previousViewConditions, $postViewConditions);
    }

    protected function addWhereFromCustomViewConditions(ZmartFilterCriteria $filterCriteria, $previousViewConditions, $postViewConditions) {
        if(!empty($previousViewConditions)) {
            $filterCriteria->setPreviousComparisonExpression($this->buildComparisonObj($previousViewConditions));
        }
        if(!empty($postViewConditions)) {
            $filterCriteria->setPostComparisonExpression($this->buildComparisonObj($postViewConditions));
        }
    }

    protected function buildComparisonObj($comparisonArray) {
        $comparisonObj = null;

        if(isset($comparisonArray['operator'])){
            if($comparisonArray['operator'] === "||" || $comparisonArray['operator'] === "&&") {
                $expressions = array();
                if($comparisonArray['operator'] === "||") {
                    foreach($comparisonArray['expressions'] as $expression) {
                        $expressions[] = $this->buildComparisonObj($expression);
                    }
                } else if($comparisonArray['operator'] === "&&") {
                    $expressions[] = $this->getBuildedValue($comparisonArray['firstValue']);
                    $expressions[] = $this->getBuildedValue($comparisonArray['secondValue']);
                }
                $comparisonObj = new LogicalExpression($comparisonArray['operator'], $expressions);
            } else {
                $operator = $comparisonArray['operator'];
                $firstValue = $this->getBuildedValue($comparisonArray['firstValue']);
                $secondValue = $this->getBuildedValue($comparisonArray['secondValue']);
                $comparisonObj = new ComparisonExpression($firstValue, $operator, $secondValue);
            }
        }

        return $comparisonObj;
    }

    protected function getBuildedValue($value) {
        if(is_array($value)) {
            $value = $this->buildComparisonObj($value);
        }
        return $value;
    }

    protected function addWhereDateRangeFromPeriod(ZmartFilterCriteria $filterCriteria, $dataSource, $period, $dateFilterColumn) {
        $dateFilterColumnToUse = null;
        if ($dateFilterColumn) {
            $dateFilterColumnToUse = $dateFilterColumn;
        } else if(isset($dataSource['dateFilterColumn']) && $dataSource['dateFilterColumn']) {
            $dateFilterColumnToUse = $dataSource['dateFilterColumn'];
        }
        if($dateFilterColumnToUse) {
            $startValue = Type::getType('date')->convertToDatabaseValue($period['start']);
            $endValue = Type::getType('date')->convertToDatabaseValue($period['end']);
            $betweenValue = array($startValue, $endValue);

            $filterCriteria->addCondition($dateFilterColumnToUse, FilterCriteria::BETWEEN, $betweenValue);
        }
    }

    public function addWhereFromRelations(ZmartFilterCriteria $filterCriteria, $dataSource, $structures) {
        if(isset($dataSource['relatedStructuresType']) && !empty($structures) && !$this->isViewPreview) {
            $structuresExternalIds = array();
            if(Utilities::isNumericArray($structures)) {
                foreach ($structures as $structure) {
                    $decodedExternalId = Utilities::decodeStructureExternalId($structure['externalId']);
                    foreach ($decodedExternalId as $key => $value) {
                        $structuresExternalIds[$key][] = $value;
                    }
                }
            } else {
                $decodedExternalId = Utilities::decodeStructureExternalId($structures['externalId']);
                foreach ($decodedExternalId as $key => $value) {
                    $structuresExternalIds[$key][] = $value;
                }
            }
            //$decodedExternalId = $structure['externalId']; //@TODO When all DataBases start use object externalId, uncomment this line
            foreach ($dataSource['relatedStructuresType'] as $relatedStructureType) {
                if($relatedStructureType['structureType'] === $structures[0]['type']) {
                    $this->buildWhereFromRelations($structuresExternalIds, $relatedStructureType['relations'], $filterCriteria);
                }
            }
        }
    }

    protected function addWhereFromDrillDownfilters(ZmartFilterCriteria $filterCriteria, $drillDownFilters, $dataSource, $view) {
        foreach ($drillDownFilters as $key => $value) {
            $drillDownColumns = $this->getDataSourceColumns($dataSource, $key, $view);
            $this->buildWhereFilterCriteria($filterCriteria, $key, $value, $drillDownColumns);
        }
    }

    protected function buildWhereFilterCriteria(&$filterCriteria, $key, $value, $drillDownColumns){

        if(!empty($drillDownColumns)) {

            if (is_array($drillDownColumns)){
                foreach ($drillDownColumns as $drillDownColumn) {
                    if(isset($drillDownColumn['isDate']) && $drillDownColumn['isDate'] ) {
                        $value = $this->convertStringToMongoDateStartEnd($value);
                        $filterCriteria->addCondition($key, FilterCriteria::BETWEEN, $value);
                    } else {
                        // Use IN operator instead of EQ because there may be drillDownFilters of an nonexistent field
                        $filter = array(null, $value);
                        if(is_array($value)){
                            $filter = $value;
                            $filter[] = null;
                        }
                        $filterCriteria->addCondition($key, FilterCriteria::IN, $filter);
                    }
                }
            }else {
                if(isset($drillDownColumns['isDate']) && $drillDownColumns['isDate'] ) {
                    $value = $this->convertStringToMongoDateStartEnd($value);
                    $filterCriteria->addCondition($key, FilterCriteria::BETWEEN, $value);
                } else {
                    // Use IN operator instead of EQ because there may be drillDownFilters of an nonexistent field
                    $filter = array(null, $value);
                    if(is_array($value)){
                        $filter = $value;
                        $filter[] = null;
                    }
                    $filterCriteria->addCondition($key, FilterCriteria::IN, $filter);
                }

            }
        }
    }

    protected function convertStringToMongoDateStartEnd($dateString) {
        $interval = Period::convertStringToMoment($dateString, $dateString);
        $interval[0] = Type::getType('date')->convertToDatabaseValue($interval[0]);
        $interval[1] = Type::getType('date')->convertToDatabaseValue($interval[1]);
        return $interval;
    }

    protected function setFilterCriteriaAggregation(ZmartFilterCriteria $filterCriteria, $facts, $dimension, $orderBy, $dataSource, $structures, $structureType, $groupByStructure, $view) {
        foreach ($facts as $fact) {
            if(Utilities::isColumnGenerated($fact) && isset($fact['function']) && $fact['name']) {
                $filterCriteria->addAggregateFunction($fact['name'], $fact['function'], $fact['name']);
            } else {
                $this->addExternalFunctions($filterCriteria, $fact);
                $this->addExternalMathOperators($filterCriteria, $fact);
            }
        }

        if(isset($structures) && Utilities::isNumericArray($structures) && !empty($structures)) {
            $structures = $structures[0];
        }
        if ($groupByStructure && ($structures || $structureType)) {
            $childStructureType = $structureType?: $this->hierarchyTypeService->getChildStructureTypeByStructureType($structures['type']);
            $this->addGroupFromChildStructureType($filterCriteria, $dataSource, $childStructureType, $view);
        } else {
            if(is_array($dimension)) {
                $filterCriteria->setGroupBy($dimension);
            } else if(is_string($dimension)) {
                $filterCriteria->addGroupBy($dimension);
            }
        }

        foreach ($orderBy as $key => $direction) {
            if(is_array($direction) && !empty($direction)){
                $key = key($direction);
                $direction = current($direction);
                $filterCriteria->addOrderBy($key, strtoupper($direction));
            } else {
                $filterCriteria->addOrderBy($key, strtoupper($direction));
            }
        }
    }

    protected function addGroupFromChildStructureType(FilterCriteria $filterCriteria, $dataSource, $childStructureType, $view) {
        if (isset($dataSource['relatedStructuresType'])) {
            $descriptionField = null;
            $columns = $this->identifyColumnsToUse($dataSource, $view);
            foreach ($columns as $dataSourceColumn) {
                if ($dataSourceColumn['name'] == $childStructureType['descriptionField']) {
                    $descriptionField = $dataSourceColumn['name'];
                }
            }
            if ($descriptionField) {
                $filterCriteria->addGroupBy($descriptionField);
            } else {
                foreach ($dataSource['relatedStructuresType'] as $relatedStructureType) {
                    if($relatedStructureType['structureType'] === $childStructureType[Manager::MONGO_KEY_COLUMN]) {
                        foreach ($relatedStructureType['relations'] as $relation) {
                            if (in_array($relation['structureColumn'], $childStructureType['fieldKey'])) {
                                $filterCriteria->addGroupBy($relation['dataSourceColumn']);
                            }
                        }
                    }
                }
            }
        }
    }

    protected function addExternalFunctions(ZmartFilterCriteria $filterCriteria, $fact) {
        if(isset($fact['functions'])) {
            $functions = array();
            foreach ($fact['functions'] as $function) {
                $content = gettype($function['content']) === 'array' ? $this->factoryMathExpression($function['content']) : $function['content'];
                $functions[] = new AggregateFunction($function['name'], $function['function'], $content);
            }
            $functions = array_merge($filterCriteria->getAggregateFunctions(), $functions);
            $filterCriteria->setAggregateFunctions($functions);
        }
    }

    protected function addExternalMathOperators(ZmartFilterCriteria $filterCriteria, $fact) {
        if(isset($fact['mathExpressions'])) {
            $mathExpressions = array();
            foreach ($fact['mathExpressions'] as $mathExpression) {
                $mathExpressions[] = array(
                    'label' => $mathExpression['name'],
                    'mathExpression' => $this->factoryMathExpression($mathExpression)
                );
            }
            $mathExpressions = array_merge($filterCriteria->getMathExpressions(), $mathExpressions);
            $filterCriteria->setMathExpressions($mathExpressions);
        }
    }

    protected function factoryMathExpression(array $mathExpression) {
        $firstValue = $this->getMathExpressionFirstValue($mathExpression);
        if(gettype($firstValue) === 'array') {
            $firstValue = $this->factoryMathExpression($firstValue);
        }
        $secondValue = $this->getMathExpressionSecondValue($mathExpression);
        if(gettype($secondValue) === 'array') {
            $secondValue = $this->factoryMathExpression($secondValue);
        }
        $operator = $this->getMathExpressionOperator($mathExpression);

        return new MathExpression($firstValue, $operator, $secondValue);
    }

    protected function getMathExpressionOperator(array $mathExpression) {
        return $mathExpression['operator'];
    }

    protected function getMathExpressionFirstValue(array $mathExpression) {
        return $mathExpression['firstValue'];
    }

    protected function getMathExpressionSecondValue(array $mathExpression) {
        return $mathExpression['secondValue'];
    }

    private function buildWhereFromRelations($decodedExternalId, $relations = array(), ZmartFilterCriteria $filterCriteria) {
        if(!$relations && $decodedExternalId){
            $filterCriteria->addCondition('notRelationZmartInExternalId', FilterCriteria::EQ, 2);;
        }
        foreach ($relations as $relation) {
            $key = $relation['dataSourceColumn'];
            $value = $decodedExternalId[$relation['structureColumn']];
            $value = array_unique($value);
            $value = array_merge(...array_map(function ($val) {
                return [(string)$val,(int)$val];
            },$value));
            $filterCriteria->addCondition($key, FilterCriteria::IN, $value);
        }
    }

    protected function getDefaultDimension($view, $viewType) {
        $dimension = null;
        if(isset($viewType['dimensionProperty'])) {
            $dimension = $view[$viewType['dimensionProperty']['name']];
        }
        return $dimension;
    }

    protected function getDefaultFacts($widget, $viewType, $facts = array()) {
        if(!empty($viewType["name"]) && $viewType["name"] == "dashboardCard"){
            if(!empty($widget["cardValue"])){
                $facts[] = $widget["cardValue"];
            }
            if(!empty($widget["cardConditions"])){
                foreach($widget["cardConditions"] as $condition){
                    if(!is_numeric($condition)){
                        $facts[] = ltrim($condition);
                    }
                }
            }
        }else if(isset($viewType['factProperties']) && is_array($viewType['factProperties'])) {
            foreach ($viewType['factProperties'] as $property) {
                if(isset($widget[$property['name']])) {
                    if(is_array($widget[$property['name']])) {
                        $facts = array_merge($facts, $widget[$property['name']]);
                    } else {
                        $facts[] = $widget[$property['name']];
                    }
                }
            }
        } else if(isset($viewType['factProperty']) && is_array($viewType['factProperty'])){
            $property = $viewType['factProperty'];
            if(is_array($widget[$property['name']])){
                $facts = $widget[$property['name']];
            }else{
                array_push($facts,$widget[$property['name']]);
            }
        }
        $this->gaugeFacts($widget,$facts);
        return $facts;
    }

    protected function gaugeFacts($widget, &$facts){
        if($widget['serieType'] == "gauge" && !empty($widget["serieMaxValue"]) && is_string($widget["serieMaxValue"])){
            array_push($facts,$widget["serieMaxValue"]);
        }
        if($widget['serieType'] == "gauge" && !empty($widget["serieRanges"]) && is_array($widget["serieRanges"])){
            foreach($widget["serieRanges"] as $range){
                if(is_string($range)){
                    array_push($facts,$range);
                }
            }
        }
    }

    public static function validateValuesInIndexes($indexes,$row1,$row2){

        foreach ($indexes as $index) {
            if($row1[$index] != $row2[$index]){
                return false;
            }
        }
        return true;

    }

    public static function getRowMatchInCrossTab(array $dataSource,array $dimensions,array $rowValue){
        foreach ($dataSource as $key => $row){
            if(self::validateValuesInIndexes($dimensions,$row,$rowValue)){
                return  $key;
            }
        }
        return null;
    }

    public static function createRowCrossTab($row,$dimensions,$rowFactValue,$rowFactLabel){
        $auxRow = array();
        foreach ($dimensions as $dimension) {
            if(isset($row[$dimension])){
                $auxRow[$dimension] = $row[$dimension];
            }else{
                $auxRow[$dimension] = 'Coluna vazia';
            }
        }
        $auxRow[$rowFactLabel] = $rowFactValue;
        return $auxRow;

    }

    public function setDataSetInCrossTab($factLabel,$factValue,$dimensions, $returnData)
    {
        $dataSource = array();
        $nameFields = array();
        foreach ($returnData as $index => $row) {
            if (isset($row[$factLabel], $row[$factValue])) {
                $rowFactLabel =  (String) $row[$factLabel];
                $rowFactName = Utilities::transliterateString($row[$factLabel]);
                $rowFactValue = $row[$factValue];
                $key = self::getRowMatchInCrossTab($dataSource, $dimensions, $row);
                if (!is_null($key)) {
                    if (isset($dataSource[$key][$rowFactName])) {
                        $dataSource[$key][$rowFactName] += $rowFactValue;
                    } else {
                        $dataSource[$key][$rowFactName] = $rowFactValue;
                    }
                } else if (!is_null($rowFactValue)) {
                    $dataSource[] = self::createRowCrossTab($row, $dimensions, $rowFactValue, $rowFactName);
                }
                if (!in_array($rowFactLabel, $nameFields)) {
                    $nameFields[] = $rowFactLabel;
                }
            }
        }
        if (isset($dataSource[0])) {
            $dataSource[0]['ZmartFields'] = $nameFields;
        }
        $returnData = $dataSource;
        return $returnData;
    }
}
